package betterwithmods.common.entity;

import betterwithmods.common.BWMBlocks;
import betterwithmods.common.blocks.mechanical.tile.TileEntityPulley;
import betterwithmods.module.GlobalConfig;
import betterwithmods.util.AABBArray;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import net.minecraft.block.Block;
import net.minecraft.block.BlockRailBase;
import net.minecraft.block.BlockRedstoneWire;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.Entity;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.util.math.Vec3i;
import net.minecraft.world.World;
import net.minecraftforge.fml.common.registry.IEntityAdditionalSpawnData;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import java.nio.charset.Charset;
import java.util.*;
import java.util.Map.Entry;

/**
 * @author mrebhan
 */

public class EntityExtendingRope extends Entity implements IEntityAdditionalSpawnData {

    private BlockPos pulley;
    private int targetY;
    private boolean up;

    private Map<Vec3i, IBlockState> blocks;
    private AABBArray blockBB;

    public EntityExtendingRope(World worldIn) {
        this(worldIn, null, null, 0);
    }

    public EntityExtendingRope(World worldIn, BlockPos pulley, BlockPos source, int targetY) {
        super(worldIn);
        this.pulley = pulley;
        this.targetY = targetY;
        if (source != null) {
            this.up = source.func_177956_o() < targetY;
            this.func_70634_a(source.func_177958_n() + 0.5, source.func_177956_o(), source.func_177952_p() + 0.5);
        }
        this.blocks = new HashMap<>();
        this.blockBB = null;
        this.func_70105_a(0.1F, 1F);
        this.field_70158_ak = true;
    }

    private static AxisAlignedBB createAABB(Vec3d part1, Vec3d part2) {
        return new AxisAlignedBB(part1.field_72450_a, part1.field_72448_b, part1.field_72449_c, part2.field_72450_a, part2.field_72448_b, part2.field_72449_c);
    }

    @Override
    protected void func_70088_a() {

    }

    @Override
    public float func_70047_e() {
        return -1;
    }

    @Override
    public void func_70107_b(double x, double y, double z) {
        if (blocks != null)
            updatePassengers(field_70163_u, y, false);
        super.func_70107_b(x, y, z);
    }

    @Override
    protected void func_70037_a(NBTTagCompound compound) {
        pulley = new BlockPos(compound.func_74762_e("PulleyX"), compound.func_74762_e("PulleyY"),
                compound.func_74762_e("PulleyZ"));
        targetY = compound.func_74762_e("TargetY");
        up = compound.func_74767_n("Up");
        if (compound.func_74764_b("BlockData")) {
            byte[] bytes = compound.func_74770_j("BlockData");
            ByteBuf buf = Unpooled.buffer(bytes.length);
            buf.writeBytes(bytes);
            blocks = deserializeBlockmap(buf);
        }
        rebuildBlockBoundingBox();
    }

    @Override
    protected void func_70014_b(NBTTagCompound compound) {
        compound.func_74768_a("PulleyX", pulley.func_177958_n());
        compound.func_74768_a("PulleyY", pulley.func_177956_o());
        compound.func_74768_a("PulleyZ", pulley.func_177952_p());
        compound.func_74768_a("TargetY", targetY);
        compound.func_74757_a("Up", up);
        ByteBuf buf = Unpooled.buffer();
        serializeBlockmap(buf, blocks);
        byte[] bytes = new byte[buf.readableBytes()];
        buf.readBytes(bytes);
        compound.func_74773_a("BlockData", bytes);

        if (GlobalConfig.debug) {
            for (int i = 0; i < bytes.length; i++) {
                if (i % 16 == 0) {
                    StringBuilder text = new StringBuilder(Integer.toHexString(i));
                    while (text.length() < 8) {
                        text.insert(0, "0");
                    }
                    System.out.print("\n" + text + ": ");
                }
                StringBuilder b = new StringBuilder(Integer.toHexString(Byte.toUnsignedInt(bytes[i])));
                while (b.length() < 2) {
                    b.insert(0, "0");
                }
                System.out.print(b);
                if (i % 2 == 1) {
                    System.out.print(' ');
                }
            }
        }
    }

    private void serializeBlockmap(ByteBuf buf, Map<Vec3i, IBlockState> blocks) {
        // TODO: Maybe add TileEntity support? It would be cool if blocks directly on top of the platform would be transported with it
        buf.writeInt(blocks.size());
        blocks.forEach((vec, state) -> {
            buf.writeInt(vec.func_177958_n());
            buf.writeInt(vec.func_177956_o());
            buf.writeInt(vec.func_177952_p());
            Block block = state != null ? state.func_177230_c() : Blocks.field_150350_a;
            ResourceLocation resourcelocation = Block.field_149771_c.func_177774_c(block);
            String blockName = resourcelocation == null ? "" : resourcelocation.toString();
            buf.writeInt(blockName.length());
            buf.writeBytes(blockName.getBytes(Charset.forName("UTF-8")));
            buf.writeByte((byte) block.func_176201_c(state));
        });
    }

    private Map<Vec3i, IBlockState> deserializeBlockmap(ByteBuf buf) {
        Map<Vec3i, IBlockState> map = new HashMap<>();
        int size = buf.readInt();
        for (int i = 0; i < size; i++) {
            Vec3i vec = new Vec3i(buf.readInt(), buf.readInt(), buf.readInt());
            int len = buf.readInt();
            byte[] bytes = new byte[len];
            buf.readBytes(bytes);
            String name = new String(bytes, Charset.forName("UTF-8"));
            int meta = buf.readByte();
            @SuppressWarnings("deprecation")
            IBlockState state = Block.func_149684_b(name).func_176203_a(meta);
            map.put(vec, state);
        }
        return map;
    }

    private void rebuildBlockBoundingBox() {
        if (blocks == null || blocks.isEmpty()) {
            this.blockBB = null;
        } else {
            List<AxisAlignedBB> bbs = new ArrayList<>();
            bbs.add(new AxisAlignedBB(0.45, 0, 0.45, 0.55, 1, 0.55)); // rope
            // bounding
            // box
            for (Vec3i vec : blocks.keySet()) {
                bbs.add(new AxisAlignedBB(vec.func_177958_n(), vec.func_177956_o(), vec.func_177952_p(), vec.func_177958_n() + 1,
                        vec.func_177956_o() + getBlockStateHeight(blocks.get(vec)), vec.func_177952_p() + 1));
            }
            this.blockBB = new AABBArray(bbs.toArray(new AxisAlignedBB[0])).func_72317_d(-0.5, 0, -0.5);
        }
    }

    @Override
    public void func_70071_h_() {
        this.field_70169_q = this.field_70165_t;
        this.field_70167_r = this.field_70163_u;
        this.field_70166_s = this.field_70161_v;

        if (up) {
            if (field_70163_u > targetY) {
                if (done())
                    return;
            }
        } else {
            if (field_70163_u < targetY) {
                if (done())
                    return;
            }
        }

        this.func_70107_b(pulley.func_177958_n() + 0.5, this.field_70163_u + (up ? 0.1 : -0.1),
                pulley.func_177952_p() + 0.5);

    }

    public void updatePassengers(double posY, double newPosY, boolean b) {
        HashSet<Entity> entitiesInBlocks = new HashSet<>();
        HashSet<Entity> passengers = new HashSet<>();
        HashMap<Entity, Double> entMaxY = new HashMap<>();

        blocks.forEach((vec, state) -> {
            if (getBlockStateHeight(state) > 0) {
                Vec3d pos = new Vec3d(pulley.func_177958_n(), posY, pulley.func_177952_p()).func_72441_c(vec.func_177958_n(), vec.func_177956_o(), vec.func_177952_p());
                func_130014_f_().func_72839_b(this,
                        createAABB(pos, pos.func_72441_c(1, getBlockStateHeight(state), 1))).forEach(e -> {
                    if (!(e instanceof EntityExtendingRope)) {
                        double targetY = pos.field_72448_b + getBlockStateHeight(state) - 0.01;
                        if (!entMaxY.containsKey(e) || entMaxY.get(e) < targetY) {
                            if ((!func_130014_f_().field_72995_K ^ e instanceof EntityPlayer) || b) {
                                entitiesInBlocks.add(e);
                                entMaxY.put(e, targetY);
                            }
                        }
                    }
                });
            }
        });

        func_130014_f_().func_72839_b(this,
                AABBArray.toAABB(this.func_174813_aQ()).func_72321_a(0, 0.5, 0).func_72317_d(0, 0.5, 0)).forEach(e -> {
            if (!(e instanceof EntityExtendingRope)
                    && (!(e instanceof EntityPlayer) || !((EntityPlayer) e).field_71075_bZ.field_75100_b))
                passengers.add(e);
        });

        passengers.forEach(e -> e.func_70091_d(null, 0, newPosY - posY, 0));
        passengers.forEach(e -> e.field_70143_R = 0);
        entitiesInBlocks.forEach(e -> e.func_70107_b(e.field_70165_t, Math.max(e.field_70163_u, entMaxY.get(e) + newPosY - posY), e.field_70161_v));
        entitiesInBlocks.forEach(e -> e.field_70160_al = false);
        entitiesInBlocks.forEach(e -> e.field_70122_E = true);
        entitiesInBlocks.forEach(e -> e.field_70181_x = Math.max(up ? 0 : -0.1, e.field_70181_x));

    }

    private double getBlockStateHeight(IBlockState blockState) {
        return (blockState == null ? 1
                : (blockState.func_177230_c() == BWMBlocks.ANCHOR ? 0.375F
                : (blockState.func_177230_c() instanceof BlockRailBase || blockState.func_177230_c() instanceof BlockRedstoneWire ? 0 : 1)));
    }

    @Override
    public boolean func_90999_ad() {
        return false;
    }

    private boolean done() {
        if (!func_130014_f_().field_72995_K) {
            TileEntity te = func_130014_f_().func_175625_s(pulley);
            if (te instanceof TileEntityPulley) {
                TileEntityPulley pulley = (TileEntityPulley) te;
                if (!pulley.onJobCompleted(up, targetY, this)) {
                    BlockPos pos = this.pulley.func_177979_c(this.pulley.func_177956_o() - targetY);
                    // rails need to be placed after all the other blocks
                    // blocks.forEach((vec, state) -> {
                    // if (!(state.getBlock() instanceof BlockRailBase)) {
                    // if (state.getBlock().)
                    // getEntityWorld().setBlockState(pos.add(vec), state, 3);
                    // }
                    // });
                    // blocks.forEach((vec, state) -> {
                    // if (state.getBlock() instanceof BlockRailBase) {
                    // getEntityWorld().setBlockState(pos.add(vec), state, 3);
                    // }
                    // });

                    int retries = 0;
                    while (!blocks.isEmpty() && retries < 10) {
                        retries++;
                        int skipped = 0;
                        for (Entry<Vec3i, IBlockState> entry : blocks.entrySet()) {
                            BlockPos blockPos = pos.func_177971_a(entry.getKey());
                            if (entry.getValue().func_177230_c().func_176196_c(func_130014_f_(), blockPos)) {
                                func_130014_f_().func_180501_a(blockPos, entry.getValue(), 3);
                                blocks.remove(entry.getKey());
                                skipped = 0;
                                break;
                            }
                            skipped++;
                        }
                        if (skipped == 0) {
                            retries = 0;
                        }
                    }

                    if (retries > 0) {
                        blocks.forEach((vec, state) -> state.func_177230_c().getDrops(func_130014_f_(), pos, state, 0).forEach(stack -> func_130014_f_()
                                .func_72838_d(new EntityItem(func_130014_f_(), field_70165_t, field_70163_u, field_70161_v, stack))));
                    }

                    updatePassengers(field_70163_u, targetY + 0.25, true);
                    return true;
                }
            }
        }
        return false;
    }

    public void addBlock(Vec3i offset, IBlockState state) {
        blocks.put(offset, state);
        rebuildBlockBoundingBox();
    }

    @Override
    public void writeSpawnData(ByteBuf buffer) {
        buffer.writeInt(pulley.func_177958_n());
        buffer.writeInt(pulley.func_177956_o());
        buffer.writeInt(pulley.func_177952_p());
        buffer.writeInt(targetY);
        buffer.writeBoolean(up);
        serializeBlockmap(buffer, blocks);
    }

    @Override
    public void readSpawnData(ByteBuf additionalData) {
        pulley = new BlockPos(additionalData.readInt(), additionalData.readInt(), additionalData.readInt());
        targetY = additionalData.readInt();
        up = additionalData.readBoolean();
        blocks = deserializeBlockmap(additionalData);
    }

    public int getTargetY() {
        return this.targetY;
    }

    public void setTargetY(int i) {
        this.targetY = i;
    }

    @Override
    public boolean func_70067_L() {
        return !this.field_70128_L;
    }

    public boolean getUp() {
        return up;
    }

    @Override
    protected boolean func_184219_q(Entity passenger) {
        return false;
    }

    public boolean isPathBlocked() {
        HashSet<BlockPos> blocked = new HashSet<>();
        blocks.forEach((vec, state) -> {
            if (blocked.isEmpty() && !up || state.func_177230_c() != BWMBlocks.ANCHOR) {
                BlockPos pos = this.pulley.func_177979_c(this.pulley.func_177956_o() - targetY).func_177971_a(vec);
                if (up)
                    pos = pos.func_177984_a();
                else
                    pos = pos.func_177977_b();

                Block b = func_130014_f_().func_180495_p(pos).func_177230_c();

                if (!(b == Blocks.field_150350_a || b.func_176200_f(func_130014_f_(), pos))) {
                    blocked.add(pos);
                }
            }
        });
        return !blocked.isEmpty();
    }

    @Override
    public boolean func_70039_c(NBTTagCompound compound) {
        return false;
    }

    public BlockPos getPulleyPosition() {
        return this.pulley;
    }

    public Map<Vec3i, IBlockState> getBlocks() {
        return blocks;
    }

    @Override
    protected void func_70105_a(float width, float height) {
        if (blockBB == null)
            super.func_70105_a(width, height);
    }

    @Override
    public void func_174826_a(AxisAlignedBB bb) {
        rebuildBlockBoundingBox();
        super.func_174826_a(blockBB != null ? blockBB.func_72317_d(this.field_70165_t, this.field_70163_u, this.field_70161_v) : bb);
    }

    public AxisAlignedBB getBlockBoundingBox(Vec3i block, IBlockState state) {
        Vec3d pos = new Vec3d(pulley.func_177958_n(), field_70163_u, pulley.func_177952_p()).func_72441_c(block.func_177958_n(), block.func_177956_o(), block.func_177952_p());
        return new AxisAlignedBB(pos, pos.func_72441_c(1, getBlockStateHeight(state), 1));
    }

    @Override
    @SideOnly(Side.CLIENT)
    public AxisAlignedBB func_70046_E() {
        return func_70067_L() ? (this.func_174813_aQ() instanceof AABBArray
                ? ((AABBArray) this.func_174813_aQ()).forEach(i -> i.func_186666_e(i.field_72337_e - 0.125))
                : this.func_174813_aQ()) : null;
    }

}
